#include <string.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <unistd.h>

#include "phero.c" 
#include "csp.c"
#include "stat.c"
#include "random.c"

float mypow(float x, int y){
	// precondition: y >= 0
	// returns x^y in O(log_2(y))
	float p;
	if (y==0) 
		return 1;
	if (y==1) 
		return x;
	p = mypow(x,y/2);
	if (y%2==0) 
		return p*p;
	return p*p*x;
}

int choose(float *cumul, int nbCand){
	float f = randfloat(); // f in [0;1[
	float total=cumul[nbCand-1];
	
	// returns 0 if f < cumul[0]/total
	// returns i if cumul[i-1]/total <= f < cumul[i]/total
	
	int left=0;
	int right=nbCand-1;
	int middle;
	while (left<right){
		// Invariant:   left=0 and f < cumul[right]/total
		//           or cumul[left-1]/total <= f < cumul[right]/total
		middle=(left+right+1)/2;
		if (f<cumul[middle-1]/total) right=middle-1;
		else if (f>=cumul[middle]/total) left=middle+1;
		else // cumul[middle-1] <= f < cumul[middle]
			return middle;
	}
	return left;
}

int buildAssignment(Tphero phero, Tcsp csp, int alpha, int beta, int *assignment){
	// computes an assignment and returns the number of violated constraints
	
	int nbAssignedVar;          // Number of var that are already assigned in assignment
	int nbCandidates;           // Number of var that are not yet assigned
	int assignedVar[csp.nbVar]; // assignedVar[i] = ith assigned var
	int candidates[csp.nbVar];  // candidates[i] = ith non assigned variable
	int nbConflicts[csp.nbVar][csp.nbVal]; 
	// for all non assigned var i, nbConflicts[i][j] = nb of violations if i=j w.r.t. assignment
	int domainSize[csp.nbVar];
	// for all non assigned var i, domainSize[i] = nb of consistent values w.r.t. assignment
	int nbViolations;            // Number of violations in assignment
	
	float total, p[csp.nbVal];
	int i, j, Xi, Vi, Vj, Xj, i_min;
	
	// Initializing data structures
	for (Xi=0; Xi<csp.nbVar; Xi++) {
		domainSize[Xi] = csp.nbVal;
		candidates[Xi] = Xi;
		assignment[Xi] = -1;
		for (Vi=0; Vi<csp.nbVal; Vi++) 
			nbConflicts[Xi][Vi] = 0;
	}
	nbViolations=0;
	
	/* Assignment of the first variable */
	Xi = next(csp.nbVar); // random choice of the first variable
	Vi = next(csp.nbVal); // random choice of the value of Xi
	assignment[Xi] = Vi;
	candidates[Xi] = csp.nbVar-1;
	nbCandidates = csp.nbVar-1;
	assignedVar[0] = Xi;
	
	for (nbAssignedVar=1; nbAssignedVar<csp.nbVar-1; nbAssignedVar++){
		
		// propagate constraints wrt the last assignment Xi=Vi
		for (i=0; i<csp.nbConstraints[Xi]; i++){
			Xj = csp.constrainedVar[Xi][i]; // Xj = ith var involved in a constraint with Xi
			if (assignment[Xj] < 0){ // Xj is not yet assigned => propagate
				for (j=0; j<csp.nbNoGoods[Xi][Xj][Vi]; j++){
					Vj = csp.noGood[Xi][Xj][Vi][j]; // (Vi,Vj) is a no good for (Xi,Xj)
					if (nbConflicts[Xj][Vj]==0) 
						domainSize[Xj]--;
					nbConflicts[Xj][Vj]++;
				}
			}
		}
		
		/* Choice of the next variable Xi to assign wrt to the min domain heuristic */
		i = next(nbCandidates); // start search from a randomly chosen variable
		i_min = i;
		for (j=i+1; j<i+nbCandidates; j++)
			if (domainSize[candidates[j%nbCandidates]]<domainSize[candidates[i_min]]) 
				i_min = j%nbCandidates;
		Xi = candidates[i_min];
		nbCandidates--;
		candidates[i_min] = candidates[nbCandidates];
		assignedVar[nbAssignedVar] = Xi;
		
		/* Choice of a value Vi for Xi w.r.t. transition probability p[Vi] */
		total = 0;
		for (Vi=0; Vi<csp.nbVal; Vi++){
			p[Vi] = total + mypow(pheroFactor(Xi,Vi,nbAssignedVar,assignedVar,assignment,phero),alpha)*
			mypow(1.0/(float)(1+nbConflicts[Xi][Vi]),beta);
			total = p[Vi];   
		}
		Vi = choose(p,csp.nbVal);
		nbViolations += nbConflicts[Xi][Vi];
		assignment[Xi]=Vi;
	};
	
	// propagate Xi=Vi on the last non assigned variable
	for (i=0; i<csp.nbConstraints[Xi]; i++){
		Xj = csp.constrainedVar[Xi][i]; // Xj = ith var involved in a constraint with Xi
		if (assignment[Xj] < 0){ // Xj is not yet assigned
			for (j=0; j<csp.nbNoGoods[Xi][Xj][Vi]; j++){
				Vj = csp.noGood[Xi][Xj][Vi][j]; // (Vi,Vj) is a no good for (Xi,Xj)
				nbConflicts[Xj][Vj]++;
			}
		}
	}
	
	// assign the last variable to a value that minimizes the number of new violations
	Xi = candidates[0];
	Vi = next(csp.nbVal);
	i = Vi;
	for (j=i+1; j<i+csp.nbVal; j++) 
		if (nbConflicts[Xi][j%csp.nbVal]<nbConflicts[Xi][Vi]) 
			Vi = j%csp.nbVal;
	assignment[Xi]=Vi;
	nbViolations += nbConflicts[Xi][Vi];
	
	return nbViolations;
}


int localSearch(int* assignment, int eval, Tcsp csp){
	// input: assignment is an assignment which violates eval constraints
	// output: improves assignment by greedy local search (using a first improvement strategy), and return the number of violated constraints
	
	int nbConflicts[csp.nbVar][csp.nbVal]; 	// nbConflicts[i][j] = nb of violations if i=j
	int violations[csp.nbVar];
	int i, j, k, Xi, Vi, Vj, Xj, nbMoves, bestVal, bestVar, gap, bestGap, lastImprovement;
	int tabu[csp.nbVar][csp.nbVal];
	
	// Initializing data structures
	for (Xi=0; Xi<csp.nbVar; Xi++){
		violations[Xi] = 0;
		for (Vi=0; Vi<csp.nbVal; Vi++){
			nbConflicts[Xi][Vi] = 0;
			tabu[Xi][Vi] = -1;
		}
	}
	for (Xi=0; Xi<csp.nbVar; Xi++){
		Vi = assignment[Xi];
		for (i=0; i<csp.nbConstraints[Xi]; i++){
			Xj = csp.constrainedVar[Xi][i]; // Xj = ith var involved in a constraint with Xi
			for (j=0; j<csp.nbNoGoods[Xi][Xj][Vi]; j++){
				Vj = csp.noGood[Xi][Xj][Vi][j]; // (Vi,Vj) is a no good for (Xi,Xj)
				nbConflicts[Xj][Vj]++;
			}
		}
	}
	for (Xi=0; Xi<csp.nbVar; Xi++) violations[Xi] = nbConflicts[Xi][assignment[Xi]];
	
	Xi = rand() % csp.nbVar;
	lastImprovement = 0;
	nbMoves = 0;
	while (1){
		// search for a move that decreases constraint violations
		bestGap = 1;
		for (i=0; ((i<csp.nbVar) && (bestGap>=0)); i++){
			Xj = (Xi+i)%csp.nbVar;
			if (violations[Xj] > 0){
				for (Vj=0; Vj<csp.nbVal; Vj++)
					if (Vj != assignment[Xj]){
						gap = nbConflicts[Xj][Vj]-nbConflicts[Xj][assignment[Xj]];
						if ((gap < bestGap) && ((gap<0) || (tabu[Xj][Vj]<lastImprovement))){
							bestGap = gap;
							bestVal = Vj;
							bestVar = Xj;
						}
					}
			}
		}
		if (bestGap>0) 
			return eval; // exit when reaching a local minima
		// Move assignment[bestVar] to bestVal
		eval += bestGap;
		violations[bestVar] += bestGap;
		for (i=0; i<csp.nbConstraints[bestVar]; i++){
			Xj = csp.constrainedVar[bestVar][i];
			for (k=0; k<csp.nbNoGoods[bestVar][Xj][assignment[bestVar]]; k++)
				nbConflicts[Xj][csp.noGood[bestVar][Xj][assignment[bestVar]][k]]--;
			for (k=0; k<csp.nbNoGoods[bestVar][Xj][bestVal]; k++)
				nbConflicts[Xj][csp.noGood[bestVar][Xj][bestVal][k]]++;
			if (!csp.isConsistent[bestVar][Xj][assignment[bestVar]][assignment[Xj]]) violations[Xj]--;
			if (!csp.isConsistent[bestVar][Xj][bestVal][assignment[Xj]]) violations[Xj]++;
		}
		assignment[bestVar] = bestVal;
		
		if (bestGap<0) lastImprovement=nbMoves;
		tabu[bestVar][bestVal] = nbMoves;
		nbMoves++;
		Xi = (bestVar+1)%csp.nbVar;
	}
	return eval;
}



int parse(int* alpha, 
		  int* beta, 
		  float* rho, 
		  int* maxCycles, 
		  int* nbAnts, 
		  float* tauMin, 
		  float* tauMax, 
		  char* fileName, 
		  int* iseed, 
		  char* strategy, 
		  char* argv[], int argc){
	// parse arguments to initialize parameters
	char ch;
	extern char* optarg;
	while ( (ch = getopt(argc, argv, "a:b:r:c:n:m:M:i:?:v:s:p:h:"))!=-1 ) {
		switch(ch) {
			case 'a': *alpha=atoi(optarg); break;
			case 'b': *beta=atoi(optarg); break;
			case 'r': *rho=atof(optarg); break;
			case 'c': *maxCycles=atoi(optarg); break;
			case 'n': *nbAnts=atoi(optarg); break;
			case 'm': *tauMin=atof(optarg); break;
			case 'M': *tauMax=atof(optarg); break;
			case 'i': strncpy(fileName, optarg, 254); break;
			case 's': *iseed=atoi(optarg); break;
			case 'p': *strategy=optarg[0]; break;
			case '?':
			default: printf("\nUsage : -a (alpha: int) -b (beta: int) -r (rho: float) ");
				printf("-c (max nb of cycles: int) -n (nb ants: int) -m (tau min: float) ");
				printf("-M (tau max: float) -i (input CSP filename) -s (seed: positive int) ");
				printf("-p (pheromonal strategy: {c=clique, v=vertex}) \n\n"); 
				return(1);
		}
	}
	return(0);
}


int main (int argc, char *argv[]) {
	
	clock_t tClock=clock(); // tclock = solution process starting time
	
	// Declare parameters and initialize them to default values
	int alpha = 2;        // Pheromone factor weight
	int beta = 8;        // Heuristic factor weight
	int nbAnts = 15;      // Number of ants
	int maxCycles = 10000; // Maximum number of cycles
	float rho = 0.02;     // Pheromone evaporation rate
	float tauMin = 0.01;  // Minimum pheromone bound
	float tauMax = 10;    // Maximum pheromone bound
	char fileName[1024];  // Name of the file that contains the CSP
	char strategy = 'v';  // strategy='c' -> Ant-SS(Clique)
	                      // strategy='v' -> Ant-SS(Vertex)
	iseed = 1;         // iseed used for generating random values
	if( parse(&alpha, &beta, &rho, &maxCycles, &nbAnts, &tauMin, &tauMax, fileName, &iseed, &strategy, argv, argc) == 1) 
		return(1);
	printf("Parameters: alpha=%i ",alpha);
	printf("beta=%d ",beta);
	printf("rho=%f ",rho);
	printf("tauMin=%f ",tauMin);
	printf("tauMax=%f ",tauMax);
	printf("nbCycles=%d ",maxCycles);
	printf("nbAnts=%d ",nbAnts);
	printf("input file name=%s ",fileName);
	printf("random seed=%d ",iseed);
	printf("pheromone strategy=%c\n",strategy);
	
	// Declaring and initializing data structures
	printf("Reading file %s and initializing data structures\n",fileName);
	int evalBestCycle, evalBest, nbCycles, i, j;
	float qty;
	Tcsp csp;
	getCSP(&csp,fileName);
	Tphero phero;
	createPhero(&phero,csp.nbVar,csp.nbVal,strategy,rho,tauMin,tauMax);
	Tstat stat;
	createTstat(csp.nbVar, csp.nbVal, &stat);
	int **ant;
	ant = (int**)calloc(nbAnts,sizeof(int*));
	for (i=0; i<nbAnts; i++) ant[i] = (int*)calloc(csp.nbVar,sizeof(int));
	int eval[nbAnts], check[nbAnts];
	int bestAssignment[csp.nbVar];
	printf("End of initialization... Starting solution process\n");
	
	evalBest = csp.nbVar * csp.nbVar;
	
	for (nbCycles=0; ((nbCycles<maxCycles) && (evalBest>0)); nbCycles++){
		evalBestCycle = csp.nbVar * csp.nbVar;
		
		// Each ant computes an assignment
		for(i=0 ; ((i<nbAnts) && (evalBest>0)) ; i++) {
			eval[i] = buildAssignment(phero, csp, alpha, beta, ant[i]);
			check[i] = update(ant[i],eval[i],&stat); 
			// check[i] = 1 if the assignment has already been constructed before
			if (check[i]==0){
				eval[i] = localSearch(ant[i],eval[i],csp);
				check[i] = update(ant[i],eval[i],&stat);
				if (check[i] == 0){
					if (eval[i] < evalBestCycle){
						evalBestCycle = eval[i];
						if (eval[i] < evalBest){
							evalBest = eval[i];
							for (j=0; j<csp.nbVar; j++) bestAssignment[j] = ant[i][j];
							printf("Number of constraint violations = %d ",evalBest);
							printf("at cycle %d (%f seconds)\n",nbCycles+1, (float)(clock()-tClock)/CLOCKS_PER_SEC);
							if (evalBest == 0){
								printSolution(csp,bestAssignment);
								return 0;
							}
						}
					}
				}
				else // resampling detected => decrease pheromone trails associated with the resampled assignment
					setTauMin(phero,ant[i]);
			}
			else // resampling detected => decrease pheromone trails associated with the resampled assignment
				setTauMin(phero,ant[i]);
		}
		
		// Pheromone updating step
		evaporate(phero);
		qty = 1.0/(float)(1+evalBestCycle-evalBest);
		for (i=0; i<nbAnts; i++)
			if ((check[i] == 0) && (eval[i] == evalBestCycle)) 
				reward(phero, ant[i], qty);
		
		// Display statistics
		if ((nbCycles+1)%100 == 0){
			printf("Statistics at cycle %d (%f seconds): ",nbCycles+1,(float)(clock()-tClock)/CLOCKS_PER_SEC);
			displayStat(stat); 
			printf("\n");
			reset(&stat);
		}
	}
	printSolution(csp,bestAssignment);
	return 0;
}
