typedef struct{
	char strategy;      // 'c' for clique; 'v' for vertex
	float **tauV;       // pheromone trails for the vertex strategy
	float ****tauC;     // pheromone trails for the clique strategy
	float rho, tauMin, tauMax; // ACO parameters
	int nbVar, nbVal;
} Tphero;

void evaporate(Tphero phero){
	int i,j,k,l;
	if (phero.strategy == 'c'){ // clique strategy
		for (i=0; i<phero.nbVar; i++)
			for (j=0; j<phero.nbVal; j++)
				for (k=i+1; k<phero.nbVar; k++)
					for (l=0; l<phero.nbVal; l++){
						phero.tauC[i][j][k][l] *= 1-phero.rho;
						if (phero.tauC[i][j][k][l] < phero.tauMin) 
							phero.tauC[i][j][k][l] = phero.tauMin;
						phero.tauC[k][l][i][j] = phero.tauC[i][j][k][l];
					}
	}
	else { // vertex strategy
		for (i=0; i<phero.nbVar; i++)
			for (j=0; j<phero.nbVal; j++){
				phero.tauV[i][j] *= 1-phero.rho;
				if (phero.tauV[i][j] < phero.tauMin) 
					phero.tauV[i][j] = phero.tauMin;
			}
	}
}

void reward(Tphero phero, int* solution, float qty){
	// add qty pheromone on the pheromone components associated with solution
	int i,j;
	if (phero.strategy == 'c'){ // clique strategy
		for (i=0; i<phero.nbVar; i++)
			for (j=i+1; j<phero.nbVar; j++){
				phero.tauC[i][solution[i]][j][solution[j]] += qty;
				if (phero.tauC[i][solution[i]][j][solution[j]] > phero.tauMax) 
					phero.tauC[i][solution[i]][j][solution[j]] = phero.tauMax;
				phero.tauC[j][solution[j]][i][solution[i]] = phero.tauC[i][solution[i]][j][solution[j]];
			}
	}
	else { // vertex strategy
		for (i=0; i<phero.nbVar; i++){
			phero.tauV[i][solution[i]] += qty;
			if (phero.tauV[i][solution[i]] > phero.tauMax) 
				phero.tauV[i][solution[i]] = phero.tauMax;
		}
	}
}

void setTauMin(Tphero phero, int* solution){
	// initialize all pheromone trails to tauMin
	int i,j;
	if (phero.strategy == 'c'){ // clique strategy
		for (i=0; i<phero.nbVar; i++)
			for (j=i+1; j<phero.nbVar; j++){
				phero.tauC[i][solution[i]][j][solution[j]] = phero.tauMin;
				phero.tauC[j][solution[j]][i][solution[i]] = phero.tauMin;
			}
	}
	else { // vertex strategy
		for (i=0; i<phero.nbVar; i++){
			phero.tauV[i][solution[i]] = phero.tauMin;
		}
	}
}


void createPhero(Tphero* phero, int nbVar, int nbVal, char strategy, float rho, float tauMin, float tauMax){
	// create and initialize data structure
	int i, j, k, l;
	phero->strategy = strategy;
	phero->rho = rho;
	phero->tauMin = tauMin;
	phero->tauMax = tauMax;
	phero->nbVar = nbVar;
	phero->nbVal = nbVal;
	if (phero->strategy == 'c'){ // clique strategy
		phero->tauC = (float****)calloc(nbVar,sizeof(float***));
		for (i=0; i<nbVar; i++){
			phero->tauC[i] = (float***)calloc(nbVal,sizeof(float**));
			for (j=0; j<nbVal; j++){
				phero->tauC[i][j] = (float**)calloc(nbVar,sizeof(float*));
				for (k=0; k<nbVar; k++){
					phero->tauC[i][j][k] = (float*)calloc(nbVal,sizeof(float));
					for (l=0; l<nbVal; l++) 
						phero->tauC[i][j][k][l] = tauMax;
				}
			}
		}
	}
	else { // vertex strategy
		phero->tauV = (float**)calloc(nbVar,sizeof(float*));
		for (i=0; i<nbVar; i++){
			phero->tauV[i] = (float*)calloc(nbVal,sizeof(float));
			for (j=0; j<nbVal; j++) 
				phero->tauV[i][j] = tauMax;
		}
	}
}


float pheroFactor(int var, int val, int nbAssignedVar, int* assignedVar, int* solution, Tphero phero){
	// returns the pheromone factor associated with var=val
	// forall i in [0..nbAssignedVar-1], assignedVar[i] = ith assigned var
	// and solution[assignedVar[i]] = value assigned to this variable
	if ((phero.strategy=='v')) return phero.tauV[var][val];
	float f = 0;
	int i;
	for (i=0; i<nbAssignedVar; i++)
		f += phero.tauC[var][val][assignedVar[i]][solution[assignedVar[i]]];
	return f;
}
