typedef struct{
	int nbVar;	// number of variables in the CSP
	int nbVal;	// number of values in each domain
	int** freq;	// forall i in 0..nbVar-1, forall j in 0..nbVal-1, freq[i][j] = number of times 
	// variable i has been assigned to value j since the last reset
	float intersections;// = sum_{i in 0..nbVar-1, j in 0..nbVal-1} freq[i][j]*(freq[i][j]-1)/2
	int nbSolutions;	// number of assignments built since the last reset
	int nbConflicts;	// number of conflicts in the hashing table 
						// -> gives a lower bound on the number of times a same assignment has been recomputed
	int total;			// sum of all violations in all assignments built since the last reset
	int prime;         
	int sizeOfMem;
	int **hashingTable;	// forall i in 0..prime-1, forall j in 0..sizeOfMem-1
						// if hashingTable[i][j] != -1, then a walk has computed a clique c such that
						// the first hash key of c if i and the second hash key of c is hashingTable[i][j]
	int *pow3;			// forall i in 0..nbVar-1, pow3[i] = 3^i (no care of capacity overflow)
	int *perm;			// perm[0..nbVar-1] is a permutation of 0..nbVar used to compute the second hash key
} Tstat;

void reset(Tstat *S){
	// reset data structures used to compute the dispersion rate and the average quality
	int i, j;
	for (i=0; i<S->nbVar; i++)
		for (j=0; j<S->nbVal; j++)
			S->freq[i][j]=0;
	S->intersections = 0;
	S->nbSolutions = 0;
	S->total = 0;
}

float getSimilarity(Tstat S){
	// return the average similarity of all assignments computed since the last reset
	if (S.nbSolutions>1){
		float f = S.intersections/((float)(S.nbSolutions-1)*(float)(S.nbSolutions));
		return 2*f/(float)(S.nbVar);
	}
	return 0;
}

void displayStat(Tstat S){
	float f = S.intersections;
	f /= (float)(S.nbSolutions-1);
	f /= (float)(S.nbSolutions);
	if (S.nbSolutions>1) 
		printf("similarity=%f ",2*f/(float)(S.nbVar));
	else printf("similarity=? ");
	printf("conflicts=%d average quality=%f",S.nbConflicts, (float)(S.total)/(float)(S.nbSolutions));
}

void createTstat(int nbVar, int nbVal, Tstat *S){
	int i,j,aux;
	
	S->nbConflicts = 0;
	S->nbVar = nbVar;
	S->freq = (int**)calloc(nbVar, sizeof(int*));
	for (i=0; i<nbVar; i++){
		S->nbVal = nbVal;
		S->freq[i] = (int*)calloc(nbVal, sizeof(int));
		for (j=0; j<nbVal; j++) S->freq[i][j] = 0;
	}
	S->intersections = 0;
	S->nbSolutions = 0;
	S->total = 0;
	S->prime = 452279;
	S->sizeOfMem = 10;
	// initialization of the hashing table: 
	// forall i in 0..prime-1, foralll j in 0..sizeOfMem-1, hashingTable[i][j]=-1
	S->hashingTable = (int**)calloc( S->prime, sizeof(int*) );
	for (i=0; i<S->prime; i++){
		S->hashingTable[i] = (int*)calloc( S->sizeOfMem, sizeof(int) );
		for (j=0; j<S->sizeOfMem; j++) S->hashingTable[i][j]=-1;
	}
	// initialization of pow3: forall i in 0..nbVar-1, pow3[i]=3^i
	S->pow3 = (int*)calloc( (S->nbVar), sizeof(int) );
	S->pow3[0]= 1;
	for (i=1; i<S->nbVar; i++) 
		S->pow3[i] = S->pow3[i-1]*3;
	// initialization of perm[0..nbVar-1] to a permutation of [0..nbVertices-1]
	S->perm = (int*)calloc( S->nbVar, sizeof(int) );
	for (i=0; i<S->nbVar; i++) S->perm[i]=i;
	for (i=0; i<S->nbVar; i++){
		j = i+rand()%(S->nbVar-i);
		aux = S->perm[i];
		S->perm[i] = S->perm[j];
		S->perm[j] = aux;
	}
}


int update(int* solution, int eval, Tstat *S){
	// update statistics wrt to solution
	int i;
	int h1 = 0;
	int h2 = 0;
	
	S->total += eval;
	S->nbSolutions++;
	for (i=0; i<S->nbVar; i++){
		S->intersections += S->freq[i][solution[i]];
		S->freq[i][solution[i]]++;
		h1 = h1+solution[i]*S->pow3[i];	
		h2 = h2+solution[i]*S->pow3[S->perm[i]];	  
	}
	if (h1<0)
		h1 = -h1;
	h1 = h1 % S->prime;	
	for (i=0; ((i<S->sizeOfMem) && (S->hashingTable[h1][i]!=h2) && (S->hashingTable[h1][i]>=0)); i++);
	if (i==S->sizeOfMem){
		printf("Memory exceeded... should realloc !\n");
		return 0;
	}
	else if (S->hashingTable[h1][i]==h2){
		S->nbConflicts++; // conflict detected
		return 1;
	}
	else{ // hashingTable[h1][i]<0) -> add (h1,h2) to the hashing Table
		S->hashingTable[h1][i]=h2;
		return 0;
	}
}

